import sys


"""Each node can be colored RED or BLACK."""
RED = "RED"
BLACK = "BLACK"


class NilNode(object):

    def __init__(self):
        self.color = "BLACK"

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self.color == other.color
        else:
            return False

# """We define NIL to be the leaf sentinel of our tree."""
NIL = NilNode()


class Node(object):

    def __init__(self, key, color="RED", left=NilNode(), right=NilNode(), p=NilNode()):
        """Constructs a single node of the red-black tree.
        Key is the key that has an ordering.
        color is RED or BLACK.
        left and right are the left and right subtrees.
        p is the parent Node.
        """
        assert color in (RED, BLACK)
        self.color, self.key, self.left, self.right, self.p = (
            color, key, left, right, p)

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return (self.color == other.color) and (self.key == other.key)
        else:
            return False

class Tree(object):

    def __init__(self, root=NilNode()):
        self.root = root


def left_rotate(T, x):
    """Left-rotates node x on tree T.

               x
              / \
             a   y
                / \
               b   g

    mutates into:

               y
              / \
             x   g
            / \
           a   b

    Used for maintaining tree balance.
    """
    assert (x.right != NIL)
    y = x.right
    x.right = y.left
    if y.left != NIL:
        y.left.p = x
    y.p = x.p
    if x.p == NIL:
        T.root = y
    elif x == x.p.left:
        x.p.left = y
    else:
        x.p.right = y
    y.left = x
    x.p = y


def right_rotate(T, x):
    """Right-rotates node x on tree T.

               x
              / \
             y   g
            / \
           a   b

    mutates into:

               y
              / \
             a   x
                / \
               b   g

    Used for maintaining tree balance.
    """
    assert (x.left != NIL)
    y = x.left
    x.left = y.right
    if y.right != NIL:
        y.right.p = x
    y.p = x.p
    if x.p == NIL:
        T.root = y
    elif x == x.p.right:
        x.p.right = y
    else:
        x.p.left = y
    y.right = x
    x.p = y


def tree_insert(tree, z):
    """Inserts node 'z' into binary tree 'tree'."""
    y = NIL
    x = tree.root
    while x != NIL:
        y = x
        if z.key < x.key:
            x = x.left
        else:
            x = x.right
    z.p = y
    if y == NIL:
        tree.root = z
    elif z.key < y.key:
        y.left = z
    else:
        y.right = z


def rb_insert(tree, x):
    """Does an insertion of 'x' into the red-black tree 'tree'.  The
    algorithm here is a little subtle, but is explained in CLR."""
    tree_insert(tree, x)
    x.color = RED
    while x != tree.root and x.p.color == RED:
        if x.p == x.p.p.left:
            y = x.p.p.right
            if y.color == RED:
                x.p.color = BLACK
                y.color = BLACK
                x.p.p.color = RED
                x = x.p.p
            else:
                if x == x.p.right:
                    x = x.p
                    left_rotate(tree, x)
                x.p.color = BLACK
                x.p.p.color = RED
                right_rotate(tree, x.p.p)
        else:
            y = x.p.p.left
            if y.color == RED:
                x.p.color = BLACK
                y.color = BLACK
                x.p.p.color = RED
                x = x.p.p
            else:
                if x == x.p.left:
                    x = x.p
                    right_rotate(tree, x)
                x.p.color = BLACK
                x.p.p.color = RED
                left_rotate(tree, x.p.p)
    tree.root.color = BLACK


def iter_tree_search(x, k):
    """
        Iteratively searches a binary tree for a particular key.
        Args:
            x: the node at which to start the search
        Returns:
            None, or the node containing the key.
    """
    while x != NIL and k != x.key:
        if k < x.key:
            x = x.left
        else:
            x = x.right
    return x


def test_search():

    with open(sys.argv[1]) as file:
        file = file.read().split("\n")
        tree = Tree()
        for data in file:
            data = data.split("\t")
            #print(data, len(data))
            data[1] = int(data[1])
            if data[0] == "I":
                rb_insert(tree, Node(data[1]))
            else:
                node = iter_tree_search(tree.root, data[1])
                if node!=NIL:
                    print(node.key, "found")
                else:
                    print(data[1], "not found")

test_search()