import argparse
import tkinter
from tkinter import filedialog
import configparser
import os
import diversity
from ngram import ngram
import glob
import bytecode_generate
import visualize_edit_distance
import logging
# String matching to list the algorithms that are going to be displayed

file_name_bytecode_map = {}
file_name_bytecode_map_readable = {}

def getAlgo(algList):
    result = [False, False, False, False, False]
    algList = algList.lower()
    algos = algList.split(",")
    for i in range(len(algos)):
        algos[i] = algos[i].strip()
    if "default" in algos:
        return ([True, True, True, True, True])
    if "jaccard" in algos:
        result[0] = True
    if "smith" in algos:
        result[1] = True
    if "gjaccard" in algos:
        result[2] = True
    if "shared fraction" in algos:
        result[3] = True
    if "average" in algos:
        result[4] = True
    return result

def getSourceDirectory(file_name, source_folders):
    path = None
    for dirname in source_folders:
        dirname = dirname.rstrip(' ')
        dirname = dirname.lstrip(' ')
        path = os.path.join(dirname, file_name)
        if os.path.exists(path):
            break
        else:
            path = None
            sub_dirs = [x[0] for x in os.walk(dirname)]
            for sub_dir in sub_dirs:
                path = os.path.join(os.path.abspath(sub_dir), file_name)
                if os.path.exists(path):
                    break
                else:
                    path = None
            if path != None:
                break
    if path == None:
        return None
    else:
        return os.path.dirname(path)

def generate_bytecode(source_file, source_folders):

    bytecode_sequence = []
    bytecode_sequence_readable = []
    not_found=set()
    print('generating bytecode for:', source_file)
    with open(source_file, "r") as infile:
        lines = infile.readlines()

    for line in lines:
        #Get name of the file and line number from trace
        file_name = line[0:line.index('(')]
        line_num = line[line.index('(')+1: line.index(')')]
        #Check if the bytecode for file has already been generated
        if file_name not in file_name_bytecode_map: #If not
            bytecode_map = {}
            bytecode_map_readable = {}
            #Get source folder for the file in the trace
            src = getSourceDirectory(file_name, source_folders)
            if src == None and file_name not in not_found:
                log.info('Unable to obtain absolute path for file name:'+ file_name+ ' in trace.')
                not_found.add(file_name)
                continue
            #Create bytecode for whole file
            name = file_name.split('.')
            instructions = bytecode_generate.createBytecode(src, name[0], name[1])
            #Store instructions corresponding to each line in a dict with key as the
            #line number value as list of bytecode instructions
            i=0
            while i < len(instructions):
                inst =[]
                inst_readable = []
                start = instructions[i].starts_line
                while True:
                    inst.append(Instruction(instructions[i].opcode, instructions[i].arg))
                    inst_readable.append(instructions[i])
                    i+=1
                    if i>=len(instructions) or instructions[i].starts_line != None:
                        break
                bytecode_map[int(start)] = inst
                bytecode_map_readable[int(start)] = inst_readable
            file_name_bytecode_map[file_name] = bytecode_map
            file_name_bytecode_map_readable[file_name] = bytecode_map_readable
        
        bytecode_map = file_name_bytecode_map[file_name]
        bytecode_map_readable = file_name_bytecode_map_readable[file_name]

        line_num=int(line_num)
        if line_num in bytecode_map:
            bytecode_sequence.extend(bytecode_map[line_num])
            bytecode_sequence_readable.extend(bytecode_map_readable[line_num])
        else:
            log.info('Bytecode for line number: '+str(line_num)+ ' in file:'+ file_name+ ' not found.')
    return bytecode_sequence, bytecode_sequence_readable


class Instruction():
    def __init__(self, opcode, arg):
        self.arg = arg
        self.opcode = opcode
    def __str__(self):
        return str(self.opcode)+" "+str(self.arg)
    def __hash__(self):
        return hash((self.arg, self.opcode))
    def __eq__(self, othr):
        return ((self.opcode, self.arg) ==
                (othr.opcode, othr.arg))

if __name__ == "__main__":

    log = logging.getLogger('Trace_Compare')
    logging.basicConfig(
        level=logging.INFO,
        format='%(asctime)s:%(levelname)s:%(message)s',
        filename="out.log",
        filemode='a'
    )
    parser = argparse.ArgumentParser("trace compare")
    parser.add_argument('-c', action='store',
                        dest='select_file',
                        help='Read folder location from INI file')

    # Result of command line arguments
    results = parser.parse_args()
    if results.select_file == None:
        # If no argument was given in the command line then
        # a dialog is displayed for user to select the config file
        root = tkinter.Tk()
        setting_file = filedialog.askopenfilename(
            parent=root,
            title='Please select a config file')
        if len(setting_file) > 0:
            print("You chose:", setting_file)
    else:
        # -c argument with a file location
        setting_file = results.select_file
    config = configparser.ConfigParser()

    if not os.path.isfile(setting_file):
        log.info('Please specify a valid configuration file')
        exit(1)

    config.read(setting_file)
    if "Settings_Compare" in config:
        settings = config['Settings_Compare']
        source = settings.get('Source', os.path.basename(setting_file))
        window_size = int(settings.get('WindowSize', 10))
        ngram_n = int(settings.get('Ngram Size', 5))
        do_winnow = settings.get('Winnow', 'True')
        al = settings.get('Algorithm', 'Default')
        original = settings.get('Original', None)
        outfile = settings.get('Outfile', None)
        resultfile = settings.get('Resultfile', None)
        visualizationfile = settings.get('Visualization', None)
        source_folders = settings.get('Source_folders', []).split(',')
        save_bytecode_flag = settings.get('Save_Bytecode', False)
        destination = settings.get('Destination', os.path.join(os.path.dirname(setting_file), "Bytecodes"))
        algoList = getAlgo(al)
    else:
        print("Selected config file does not contain relevant options")
        exit(1)
    log.info("Source: "+ source)
    log.info("Algorithm: "+ al)
    log.info("Window Size:"+ str(window_size))
    log.info("N Gram Size:"+ str(ngram_n))
    log.info("Winnowing enabled:"+ do_winnow)
    log.info("Original: "+ str(original))
    log.info("Out file: "+ outfile)
    log.info("Result File: "+ str(resultfile))
    log.info("Visualization File: "+ visualizationfile)
    log.info("Source Folders:"+ str(source_folders))
    log.info("Save Bytecode: "+str(save_bytecode_flag))

    if save_bytecode_flag and not os.path.exists(destination):
        os.makedirs(destination)

    files = glob.glob(source + "/*.txt")

    bytecode_files =[]
    trace_bytecode_map = {}

    for file in files:
        bytecode_sequence, bytecode_sequence_readable = generate_bytecode(file, source_folders)
        trace_bytecode_map[file] = bytecode_sequence
        filename = os.path.basename(file)
        ext = filename.split(".")[1]

        if save_bytecode_flag:
            bytecode_file_name = os.path.join(destination, filename.split(".")[0] + "_bytecode." + ext)
            bytecode_files.append(bytecode_file_name)
            with open(bytecode_file_name, "w") as out:
                for bytecode in bytecode_sequence_readable:
                    out.write(str(bytecode) + "\n")

    print('comparing traces')
    if al == "EditDistance":
        visualize_edit_distance.editDistanceSeq(trace_bytecode_map, visualizationfile, outfile)
    else:
        dict_of_sets = ngram().getFingerprint(
            trace_bytecode_map, window_size, ngram_n, do_winnow, original)
        diversity.diversity(dict_of_sets, type, algoList,
                            outfile, original, visualizationfile)